#ifndef __HESTON_MC_H
#define __HESTON_MC_H

#include "dco_type.h"

#include <cmath>
#include <vector>
#include "option.h"

// The HestonEuler class stores the necessary information
// for creating the volatility and spot paths based on the
// Heston Stochastic Volatility model. 
class HestonEuler {
 private:
  Option* pOption;
  dco_type<double> kappa;
  dco_type<double> theta;
  dco_type<double> xi;
  dco_type<double> rho;

 public:
  HestonEuler(Option* _pOption, 
              dco_type<double> _kappa, dco_type<double> _theta, 
              dco_type<double> _xi, dco_type<double> _rho);
  virtual ~HestonEuler();

  // Calculate the volatility path
  void calc_vol_path(const std::vector<dco_type<double>>& vol_draws, 
                     std::vector<dco_type<double>>& vol_path);

  // Calculate the asset price path
  void calc_spot_path(const std::vector<dco_type<double>>& spot_draws, 
                      const std::vector<dco_type<double>>& vol_path, 
                      std::vector<dco_type<double>>& spot_path);
};

#endif
