#include "dco.hpp"

#include <iostream>
#include <cmath>

template<typename T>
void newton(T& x, const T& p, const float& eps) {
  do { x=x-(x*x-p)/(2*x); } while (fabs(x*x-p)>eps); 
}

template<typename T>
void adjoint_newton(typename dco::ga1s<T>::external_adjoint_object_t *D) {
  const T &x_v = D->template read_data<T>();
  T xa=D->get_output_adjoint();
  T pa=xa/(2*x_v);
  D->increment_input_adjoint(pa);
}

template<typename T>
void augmented_primal_newton(typename dco::ga1s<T>::type &x, 
	const typename dco::ga1s<T>::type &p, const float& eps) {
  using DCO_M=typename dco::ga1s<T>;
  using DCO_EA=typename DCO_M::external_adjoint_object_t;
  DCO_EA *D=DCO_M::global_tape->template create_callback_object<DCO_EA>();
  T p_v=D->register_input(p);
  T x_v=dco::value(x);
  newton(x_v,p_v,eps);
  x=D->register_output(x_v);
  D->write_data(x_v);
  DCO_M::global_tape->template insert_callback<DCO_EA>(adjoint_newton<T>,D);
}

template<typename T>
void dxdp_a(T& x_v, const T& p_v, const float& eps, T& dxdp) {
  using DCO_M=typename dco::ga1s<T>;
  using DCO_T=typename DCO_M::type;
  using DCO_TT=typename DCO_M::tape_t;
  DCO_M::global_tape=DCO_TT::create();
  DCO_T x=x_v,p=p_v;
  DCO_M::global_tape->register_variable(p);
  x=pow(x,2);
  // newton(x,p,eps); // algorithmic adjoint
  augmented_primal_newton<T>(x,p,eps); // symbolic adjoint
  x=exp(-x);
  x_v=dco::value(x);
  DCO_M::global_tape->register_output_variable(x);
  dco::derivative(x)=1;
  std::cerr << dco::size_of(DCO_M::global_tape) << "B" << std::endl;
  DCO_M::global_tape->interpret_adjoint();
  dxdp=dco::derivative(p);
  DCO_TT::remove(DCO_M::global_tape);
}

int main() {
  using T=double; 
  T p=2,x=10,dxdp;
  const float eps=1e-12;
  dxdp_a(x,p,eps,dxdp);
  std::cout << "x=" << x << std::endl;
  std::cout << "dxdp=" << dxdp << std::endl;
  return 0;
}
