#include<iostream>
#include<cmath>
#include<stack>

template<typename T>
void newton_a(T& x_v, T& x_a, const T& p_v, T& p_a, const T& eps, bool a=false) {
  static std::stack<T> tbr_T; // to be recorded
  static int i=0; // iteration counter
  static T y=0; // persistent primal result
  if (!a) { // augmented primal
    do { 
      tbr_T.push(x_v);
      x_v-=(x_v*x_v-p_v)/(2*x_v);
      i++;
    } while (fabs(x_v*x_v-p_v)>eps);
    y=x_v;
  } else { // adjoint
    for (int j=0;j<i;j++) {
      x_v=tbr_T.top(); tbr_T.pop();
      p_a+=x_a/(2*x_v);
      x_a-=(3./4.+p_v/(4*x_v*x_v))*x_a;
    }
    x_v=y;
  }
}

int main() {
  using T=double;
  T p_v=2, x_v=10, p_a=0, x_a=1;
  const T eps=1e-12;
  std::stack<T> tbr_T;

  tbr_T.push(x_v);
  x_v=pow(x_v,2);
  newton_a(x_v,x_a,p_v,p_a,eps);
  x_v=exp(-x_v);
  std::cout << "x=" << x_v << std::endl;

  x_a=-x_v*x_a;
  newton_a(x_v,x_a,p_v,p_a,eps,true);
  x_v=tbr_T.top(); tbr_T.pop();
  x_a=2*x_v*x_a;
  std::cout << "dxdp=" << p_a << std::endl;
  // std::cout << "dxdx_0=" << x_a << std::endl; // vanishes

  return 0;
}
