#include "f12.h"
#include "dco.hpp"
#include <random>

template <int VS, typename T, typename Tp, typename To=float>
T f(int np, int nb, int ns, int nc, T x, const std::vector<std::vector<Tp>>& p) {
  assert(VS>=2*ns/nc+1);
  using namespace std;
  using Tb=typename dco::mode<T>::value_t;
  default_random_engine g; normal_distribution<To> d(0,1);
  T y=0, x_in=x; To dt=1./ns;
  for (int ib=0;ib<np;ib+=np/nb)
    for (int i=ib;i<ib+np/nb;++i) {
      x=x_in;
      for (int jc=0;jc<ns;jc+=ns/nc) {
        typename dco::gt1v<Tb,VS>::type x_t=dco::value(x); 
        dco::derivative(x_t)[0]=1; 
        std::vector<typename dco::gt1v<Tb,VS>::type> p0_t(ns/nc), p1_t(ns/nc);
        for (int j=jc,k=1;j<jc+ns/nc;++j,k+=2) {
          typename dco::gt1v<Tb,VS>::type p0j_t=dco::value(p[0][j]), p1j_t=dco::value(p[1][j]);
          dco::derivative(p0j_t)[k]=1; dco::derivative(p1j_t)[k+1]=1;
          x_t+=dt*f1(x_t,p0j_t)+f2(x_t,p1j_t)*sqrt(dt)*d(g);
        }
        // record local linearization
        x=dco::derivative(x_t)[0]*x;
        for (int j=jc,k=1;j<jc+ns/nc;++j,k+=2) 
          x+=dco::derivative(x_t)[k]*p[0][j]+dco::derivative(x_t)[k+1]*p[1][j]; 
        dco::value(x)=dco::value(x_t); // correct primal result
      }
    y+=x;
  }
  return pow(y/np,2);
}

int main(int argc, char* argv[]) {
  assert(argc==5);
  int np=std::stoi(argv[1]), nb=std::stoi(argv[2]);
  assert(np>0&&nb>0&&np%nb==0);
  int ns=std::stoi(argv[3]), nc=std::stoi(argv[4]);
  assert(ns>0&&nc>0&&ns%nc==0);
  const int VS=201;
  using Tbb=double; std::cout.precision(std::numeric_limits<Tbb>::digits10);
  using Tb=dco::gt1s<Tbb>::type; using Tp=dco::ga1s<Tb>::type; using T=Tp;
  T x=1,y; std::vector<std::vector<Tp>> p(2,std::vector<Tp>(ns,1));
  dco::ga1s<Tb>::global_tape=dco::ga1s<Tb>::tape_t::create(); 
  dco::ga1s<Tb>::global_tape->register_variable(x); 
  for (int i=0;i<ns;i++) { 
    dco::ga1s<Tb>::global_tape->register_variable(p[0][i]); 
    dco::ga1s<Tb>::global_tape->register_variable(p[1][i]); 
  }
  dco::ga1s<Tb>::tape_t::position_t pos=dco::ga1s<Tb>::global_tape->get_position();
  // d(dy/d(x,p))dx
  dco::derivative(dco::value(x))=1; 
  y=f<VS>(np,nb,ns,nc,x,p); 
  dco::value(dco::derivative(y))=1; 
  dco::ga1s<Tb>::global_tape->interpret_adjoint_and_reset_to(pos);
  std::cout << "d(dy/dx)/dx=" << dco::derivative(dco::derivative(x)) << std::endl;
  for (int j=0;j<ns;++j) {
    std::cout << "d(dy/dp[0][" << j << "])/dx=" << dco::derivative(dco::derivative(p[0][j])) << std::endl;
    dco::derivative(dco::value(p[0][j]))=0; dco::derivative(p[0][j])=0;
    std::cout << "d(dy/dp[1][" << j << "])/dx=" << dco::derivative(dco::derivative(p[1][j])) << std::endl;
    dco::derivative(dco::value(p[1][j]))=0; dco::derivative(p[1][j])=0;
  }
  dco::derivative(dco::value(x))=0; dco::derivative(x)=0; 
  // d(dy/d(x,p))dp
  for (int i=0;i<ns;++i) {
    dco::derivative(dco::value(p[0][i]))=1;
    y=f<VS>(np,nb,ns,nc,x,p); 
    dco::value(dco::derivative(y))=1; 
    dco::ga1s<Tb>::global_tape->interpret_adjoint_and_reset_to(pos);
    std::cout << "d(dy/dp[0][" << i << "])/dx=" << dco::derivative(dco::derivative(x)) << std::endl;
    dco::derivative(dco::value(x))=0; dco::derivative(x)=0; 
    for (int j=0;j<ns;++j) {
      std::cout << "d(dy/dp[0][" << i << "])/dp[0][" << j << "]=" << dco::derivative(dco::derivative(p[0][j])) << std::endl;
      dco::derivative(dco::value(p[0][j]))=0; dco::derivative(p[0][j])=0;
      std::cout << "d(dy/dp[0][" << i << "])/dp[1][" << j << "]=" << dco::derivative(dco::derivative(p[1][j])) << std::endl;
      dco::derivative(dco::value(p[1][j]))=0; dco::derivative(p[1][j])=0;
    }
    dco::derivative(dco::value(p[0][i]))=0;
    dco::derivative(dco::value(p[1][i]))=1;
    y=f<VS>(np,nb,ns,nc,x,p); 
    dco::value(dco::derivative(y))=1; 
    dco::ga1s<Tb>::global_tape->interpret_adjoint_and_reset_to(pos);
    std::cout << "d(dy/dp[1][" << i << "])/dx=" << dco::derivative(dco::derivative(x)) << std::endl;
    dco::derivative(dco::value(x))=0; dco::derivative(x)=0; 
    for (int j=0;j<ns;++j) {
      std::cout << "d(dy/dp[1][" << i << "])/dp[0][" << j << "]=" << dco::derivative(dco::derivative(p[0][j])) << std::endl;
      dco::derivative(dco::value(p[0][j]))=0; dco::derivative(p[0][j])=0;
      std::cout << "d(dy/dp[1][" << i << "])/dp[1][" << j << "]=" << dco::derivative(dco::derivative(p[1][j])) << std::endl;
      dco::derivative(dco::value(p[1][j]))=0; dco::derivative(p[1][j])=0;
    }
    dco::derivative(dco::value(p[1][i]))=0;
  }
  std::cout << "y=" << dco::value(dco::value(y)) << std::endl;
  return 0;
}

