#include "f.h"
#include "dco.hpp"
#include <random>

template <int VS, typename T, typename Tp, typename To=float>
T g(int np, int nb, int ns, int nc, T x, const std::vector<std::vector<Tp>>& p) {
  assert(VS>=2*ns+1);
  using Tb=typename dco::mode<T>::value_t;
  using namespace std;
  default_random_engine g; normal_distribution<To> d(0,1);
  T x_in=x; To dt=1./ns;
  dco::mode<T>::global_tape->switch_to_passive();
  T y=f(np,ns,x,p);
  typename dco::mode<T>::value_t aux=2.*sqrt(dco::value(y))/np;
  dco::mode<T>::global_tape->switch_to_active();
  auto pos=dco::mode<T>::global_tape->get_position();
  // seed p_t once
  std::vector<std::vector<typename dco::gt1v<Tb,VS>::type>> p_t(2,std::vector<typename dco::gt1v<Tb,VS>::type>(ns));
  for (int j=0,k=1;j<ns;++j,k+=2) {
    p_t[0][j]=dco::value(p[0][j]); 
    dco::derivative(p_t[0][j])[k]=1;
    p_t[1][j]=dco::value(p[1][j]);
    dco::derivative(p_t[1][j])[k+1]=1;
  }
  for (int ib=0;ib<np;ib+=np/nb) {
    T x_s=0;
    typename dco::gt1v<Tb,VS>::type x_s_t=0;
    for (int i=ib;i<ib+np/nb;++i) {
      x=x_in;
      typename dco::gt1v<Tb,VS>::type x_t=dco::value(x); 
      dco::derivative(x_t)[0]=1;
      for (int jc=0;jc<ns;jc+=ns/nc) 
        for (int j=jc;j<jc+ns/nc;++j)
          x_t+=dt*f1(x_t,p_t[0][j])+f2(x_t,p_t[1][j])*sqrt(dt)*d(g);
      x_s_t+=x_t;
    }
    x_s=dco::derivative(x_s_t)[0]*x;
    for (int j=0,k=1;j<ns;++j,k+=2)
      x_s+=dco::derivative(x_s_t)[k]*p[0][j]+dco::derivative(x_s_t)[k+1]*p[1][j];
    dco::derivative(x_s)=aux; // crucial: seed needs to be available!
    dco::mode<T>::global_tape->interpret_adjoint_and_reset_to(pos);
    dco::value(x_s)+=dco::value(x_s_t);
  }
  return y;
}

int main(int argc, char* argv[]) {
  assert(argc==5);
  int np=std::stoi(argv[1]), nb=std::stoi(argv[2]);
  assert(np>0&&nb>0&&np%nb==0);
  int ns=std::stoi(argv[3]), nc=std::stoi(argv[4]);
  assert(ns>0&&nc>0&&ns%nc==0);
  using Tb=double; std::cout.precision(std::numeric_limits<Tb>::digits10); 
  using Tp=dco::ga1s<Tb>::type; using T=Tp;
  T x=1.; std::vector<std::vector<Tp>> p(2,std::vector<Tp>(ns,1));
  dco::tape_options o; o.set_blob_size_in_mbyte(4);
  dco::ga1s<Tb>::global_tape=dco::ga1s<Tb>::tape_t::create(o); 
  dco::ga1s<Tb>::global_tape->register_variable(x); 
  for (int i=0;i<ns;i++) { 
    dco::ga1s<Tb>::global_tape->register_variable(p[0][i]); 
    dco::ga1s<Tb>::global_tape->register_variable(p[1][i]); 
  }
  T y=g<401>(np,nb,ns,nc,x,p); 
  std::cout << "dy/dx=" << dco::derivative(x) << '\n';
  for (int i=0;i<ns;i++) 
    std::cout << "dy/dp[0][" << i << "]=" << dco::derivative(p[0][i]) << '\n'
              << "dy/dp[1][" << i << "]=" << dco::derivative(p[1][i]) << std::endl;
  dco::ga1s<Tb>::tape_t::remove(dco::ga1s<Tb>::global_tape); // free tape memory
  std::cout << "y=" << dco::value(y) << std::endl;
  return 0;
}

