#include "f12.h"
#include "dco.hpp"
#include <random>

template <typename T, typename Tp, typename To>
void adjoint_chunk(typename dco::mode<T>::external_adjoint_object_t *EAc) {
  int ns=EAc->template read_data<int>(); int nc=EAc->template read_data<int>();
  int jc=EAc->template read_data<int>(); To dt=EAc->template read_data<To>();
  T x_c=EAc->template read_data<T>();
  dco::mode<T>::global_tape->register_variable(x_c);
  std::vector<std::vector<Tp>>* p_p=EAc->template read_data<std::vector<std::vector<Tp>>*>();
  typename dco::mode<T>::tape_t::position_t pos=dco::mode<T>::global_tape->get_position();
  T x=x_c;
  for (int j=jc;j<jc+ns/nc;++j) {
    To Z=EAc->template read_data<To>();
    x+=dt*f1(x,(*p_p)[0][j])+f2(x,(*p_p)[1][j])*sqrt(dt)*Z;
  }
  dco::derivative(x)=EAc->get_output_adjoint();
  dco::mode<T>::global_tape->interpret_adjoint_and_reset_to(pos);
  EAc->increment_input_adjoint(dco::derivative(x_c));
}

template <typename T, typename Tp, typename To>
void adjoint_bucket(typename dco::mode<T>::external_adjoint_object_t *EAb) {
  typename dco::mode<T>::tape_t::position_t pos=dco::mode<T>::global_tape->get_position();
  int ib=EAb->template read_data<int>(); int np=EAb->template read_data<int>();
  int nb=EAb->template read_data<int>(); int ns=EAb->template read_data<int>();
  int nc=EAb->template read_data<int>(); To dt=EAb->template read_data<To>();
  T x_c=EAb->template read_data<T>(); T y=EAb->template read_data<T>(); 
  std::vector<std::vector<Tp>>* p_p=EAb->template read_data<std::vector<std::vector<Tp>>*>(); 
  for (int i=ib;i<ib+np/nb;++i) {
    T x=x_c;
    for (int jc=0;jc<ns;jc+=ns/nc) {
      typename dco::mode<T>::external_adjoint_object_t* EAc=dco::mode<T>::global_tape->template create_callback_object<typename dco::mode<T>::external_adjoint_object_t>();
      EAc->write_data(ns); EAc->write_data(nc); EAc->write_data(jc);
      EAc->write_data(dt); EAc->write_data(x); EAc->write_data(p_p);
      EAc->register_input(x);
      dco::mode<T>::global_tape->switch_to_passive();
      for (int j=jc;j<jc+ns/nc;++j) {
        To Z=EAb->template read_data<To>(); EAc->write_data(Z);
        x+=dt*f1(x,(*p_p)[0][j])+f2(x,(*p_p)[1][j])*sqrt(dt)*Z;
      }
      dco::mode<T>::global_tape->switch_to_active();
      x=EAc->register_output(dco::value(x));
      dco::mode<T>::global_tape->template insert_callback<typename dco::mode<T>::external_adjoint_object_t>(adjoint_chunk<T,Tp,To>,EAc);
    }
    y+=x;
  }
  dco::derivative(y)=EAb->get_output_adjoint();
  dco::mode<T>::global_tape->interpret_adjoint_and_reset_to(pos);
}

template <typename T, typename Tp, typename To=float>
T f(int np, int nb, int ns, int nc, T x, const std::vector<std::vector<Tp>>& p) {
  using namespace std;
  default_random_engine g; normal_distribution<To> d(0,1);
  T y=0, x_c=x; 
  To dt=1./ns;
  for (int ib=0;ib<np;ib+=np/nb) {
    typename dco::mode<T>::external_adjoint_object_t* EAb=dco::mode<T>::global_tape->template create_callback_object<typename dco::mode<T>::external_adjoint_object_t>();
    EAb->write_data(ib); EAb->write_data(np); EAb->write_data(nb);
    EAb->write_data(ns); EAb->write_data(nc); EAb->write_data(dt);
    EAb->write_data(x_c); EAb->write_data(y);
    EAb->write_data(&p);
    dco::mode<T>::global_tape->switch_to_passive();
    for (int i=ib;i<ib+np/nb;++i) {
      x=x_c;
      for (int jc=0;jc<ns;jc+=ns/nc) 
        for (int j=jc;j<jc+ns/nc;++j) {
          To Z=d(g); EAb->write_data(Z);
          x+=dt*f1(x,p[0][j])+f2(x,p[1][j])*sqrt(dt)*Z;
        }
      y+=x;
    } 
    dco::mode<T>::global_tape->switch_to_active();
    y=EAb->register_output(dco::value(y));
    dco::mode<T>::global_tape->template insert_callback<typename dco::mode<T>::external_adjoint_object_t>(adjoint_bucket<T,Tp,To>,EAb);
  }
  return pow(y/np,2);
}

int main(int argc, char* argv[]) {
  assert(argc==5);
  int np=std::stoi(argv[1]), nb=std::stoi(argv[2]);
  assert(np>0&&nb>0&&np%nb==0);
  int ns=std::stoi(argv[3]), nc=std::stoi(argv[4]);
  assert(ns>0&&nc>0&&ns%nc==0);
  using Tbb=double; std::cout.precision(std::numeric_limits<Tbb>::digits10);
  using Tb=dco::gt1s<Tbb>::type; using Tp=dco::ga1s<Tb>::type; using T=Tp;
  T x=1,y; std::vector<std::vector<Tp>> p(2,std::vector<Tp>(ns,1));
  dco::ga1s<Tb>::global_tape=dco::ga1s<Tb>::tape_t::create(); 
  dco::ga1s<Tb>::global_tape->register_variable(x); 
  for (int i=0;i<ns;i++) { 
    dco::ga1s<Tb>::global_tape->register_variable(p[0][i]); 
    dco::ga1s<Tb>::global_tape->register_variable(p[1][i]); 
  }
  dco::ga1s<Tb>::tape_t::position_t pos=dco::ga1s<Tb>::global_tape->get_position();
  // d(dy/d(x,p))dx
  dco::derivative(dco::value(x))=1; 
  y=f(np,nb,ns,nc,x,p); 
  dco::value(dco::derivative(y))=1; 
  dco::ga1s<Tb>::global_tape->interpret_adjoint_and_reset_to(pos);
  std::cout << "d(dy/dx)/dx=" << dco::derivative(dco::derivative(x)) << std::endl;
  for (int j=0;j<ns;++j) {
    std::cout << "d(dy/dp[0][" << j << "])/dx=" << dco::derivative(dco::derivative(p[0][j])) << std::endl;
    dco::derivative(dco::value(p[0][j]))=0; dco::derivative(p[0][j])=0;
    std::cout << "d(dy/dp[1][" << j << "])/dx=" << dco::derivative(dco::derivative(p[1][j])) << std::endl;
    dco::derivative(dco::value(p[1][j]))=0; dco::derivative(p[1][j])=0;
  }
  dco::derivative(dco::value(x))=0; dco::derivative(x)=0; 
  // d(dy/d(x,p))dp
  for (int i=0;i<ns;++i) {
    dco::derivative(dco::value(p[0][i]))=1;
    y=f(np,nb,ns,nc,x,p); 
    dco::value(dco::derivative(y))=1; 
    dco::ga1s<Tb>::global_tape->interpret_adjoint_and_reset_to(pos);
    std::cout << "d(dy/dp[0][" << i << "])/dx=" << dco::derivative(dco::derivative(x)) << std::endl;
    dco::derivative(dco::value(x))=0; dco::derivative(x)=0; 
    for (int j=0;j<ns;++j) {
      std::cout << "d(dy/dp[0][" << i << "])/dp[0][" << j << "]=" << dco::derivative(dco::derivative(p[0][j])) << std::endl;
      dco::derivative(dco::value(p[0][j]))=0; dco::derivative(p[0][j])=0;
      std::cout << "d(dy/dp[0][" << i << "])/dp[1][" << j << "]=" << dco::derivative(dco::derivative(p[1][j])) << std::endl;
      dco::derivative(dco::value(p[1][j]))=0; dco::derivative(p[1][j])=0;
    }
    dco::derivative(dco::value(p[0][i]))=0;
    dco::derivative(dco::value(p[1][i]))=1;
    y=f(np,nb,ns,nc,x,p); 
    dco::value(dco::derivative(y))=1; 
    dco::ga1s<Tb>::global_tape->interpret_adjoint_and_reset_to(pos);
    std::cout << "d(dy/dp[1][" << i << "])/dx=" << dco::derivative(dco::derivative(x)) << std::endl;
    dco::derivative(dco::value(x))=0; dco::derivative(x)=0; 
    for (int j=0;j<ns;++j) {
      std::cout << "d(dy/dp[1][" << i << "])/dp[0][" << j << "]=" << dco::derivative(dco::derivative(p[0][j])) << std::endl;
      dco::derivative(dco::value(p[0][j]))=0; dco::derivative(p[0][j])=0;
      std::cout << "d(dy/dp[1][" << i << "])/dp[1][" << j << "]=" << dco::derivative(dco::derivative(p[1][j])) << std::endl;
      dco::derivative(dco::value(p[1][j]))=0; dco::derivative(p[1][j])=0;
    }
    dco::derivative(dco::value(p[1][i]))=0;
  }
  std::cout << "y=" << dco::value(dco::value(y)) << std::endl;
  return 0;
}

