#include "f12.h" 
#include "dco.hpp" 
#include <random>

template <typename T, typename Tp, typename To=float>
T f(int np, int ns, T& x, const std::vector<std::vector<Tp>>& p) {
  using namespace std;
  default_random_engine g; normal_distribution<To> d(0,1);
  T y=0, x_in=x; To dt=1./ns;
  for (int i=0;i<np;++i) {
    x=x_in;
    for (int j=0;j<ns;++j) 
      x+=dt*f1(x,p[0][j])+f2(x,p[1][j])*sqrt(dt)*d(g);
    y+=x;
  }
  return pow(y/np,2);
}


int main(int argc, char* argv[]) {
  assert(argc==3); 
  int np=std::stoi(argv[1]), ns=std::stoi(argv[2]); assert(np>0&&ns>0);
  using Tb=double; 
  using Tp=dco::ga1s<Tb>::type; using T=Tp;
  T x_in=1.; std::vector<std::vector<Tp>> p(2,std::vector<Tp>(ns,1));
  dco::ga1s<Tb>::global_tape=dco::ga1s<Tb>::tape_t::create();
  dco::ga1s<Tb>::global_tape->register_variable(x_in); 
  T x=x_in; // read-only independent
  for (int i=0;i<ns;i++) { 
    dco::ga1s<Tb>::global_tape->register_variable(p[0][i]); 
    dco::ga1s<Tb>::global_tape->register_variable(p[1][i]); 
  }
  T y=f(np,ns,x,p); 
  dco::ga1s<Tb>::global_tape->register_output_variable(y); 
  dco::derivative(y)=1; 
  dco::ga1s<Tb>::global_tape->interpret_adjoint(); 
  std::cout.precision(std::numeric_limits<Tb>::digits10); 
  std::cout << "dy/dx=" << dco::derivative(x_in) << '\n';
  for (int i=0;i<ns;i++) 
    std::cout << "dy/dp[0][" << i << "]=" << dco::derivative(p[0][i]) << '\n'
              << "dy/dp[1][" << i << "]=" << dco::derivative(p[1][i]) << std::endl;
  dco::ga1s<Tb>::tape_t::remove(dco::ga1s<Tb>::global_tape); 
  std::cout << "y=" << dco::value(y) << std::endl;
  return 0;
}
