#include "f.h"
#include "dco.hpp"

template<typename Tb, typename Tp>
Tb minimize(int np, int ns, double eps, Tb x_v, const std::vector<std::vector<Tp>>& p) {
  using namespace std;
  using T=typename dco::ga1sm<Tb>::type; 
  T x=x_v,y;
  typename dco::ga1sm<Tb>::tape_t* tape=dco::ga1sm<Tb>::tape_t::create(); 
  tape->register_variable(x); 
  typename dco::ga1sm<Tb>::tape_t::position_t pos=tape->get_position();
  do {
    tape->zero_adjoints();
    y=f(np,ns,x,p); 
    tape->register_output_variable(y); 
    dco::derivative(y)=1;
    tape->interpret_adjoint_and_reset_to(pos);
    dco::value(x)-=0.1*dco::derivative(x);
  }
  while (fabs(dco::derivative(x))>eps);
  dco::ga1sm<Tb>::tape_t::remove(tape); 
  return dco::value(x);
}

int main(int argc, char* argv[]) {
  assert(argc==4); 
  int np=std::stoi(argv[1]), ns=std::stoi(argv[2]); assert(np>0&&ns>0);
  using Tb=double; // base type
  Tb eps=std::atof(argv[3]); assert(eps>0);
  using Tp=dco::ga1sm<Tb>::type; using T=Tp;
  T x=1.; std::vector<std::vector<Tp>> p(2,std::vector<Tp>(ns,1));
  dco::ga1sm<Tb>::tape_t *tape=dco::ga1s<Tb>::tape_t::create(); // "touch" tape
  tape->register_variable(x); 
  for (int i=0;i<ns;i++) { // register independents
    tape->register_variable(p[0][i]); 
    tape->register_variable(p[1][i]); 
  }
  T y=minimize(np,ns,eps,x,p); // overloaded f records tape
  tape->register_output_variable(y); // register dependent
  dco::derivative(y)=1; // seed
  tape->interpret_adjoint(); // interpret tape
  std::cout.precision(std::numeric_limits<Tb>::digits10); 
  for (int i=0;i<ns;i++) 
    std::cout << "dx/dp[0][" << i << "]=" << dco::derivative(p[0][i]) << '\n'
              << "dx/dp[1][" << i << "]=" << dco::derivative(p[1][i]) << std::endl;
  dco::ga1sm<Tb>::tape_t::remove(tape); // free tape memory
  std::cout << "x=" << dco::value(y) << std::endl;
  return 0;
}
