// Uwe.Naumann@nag.com

#include <cmath>
#include <stack>
#include <iostream>
using namespace std;

template<typename T>
void ah(const T& x, T& ax, T& h_, T& ah_, bool adjoint=false) {
  if (!adjoint) 
    h_=pow(x,2); 
  else { 
    ax+=2*x*ah_; 
    ah_=0; 
  }
}

template<typename T>
void ag(const T& x, T& ax, const T& h, T& ah, T& g_, T& ag_, bool adjoint=false) { 
  static stack<T> tbr;
  if (!adjoint) { 
    T v=x*h; 
    g_=sin(v); 
    tbr.push(v); 
  }
  else { 
    T v=tbr.top(); tbr.pop(); 
    T av=cos(v)*ag_; ag_=0; 
    ax+=h*av; ah+=x*av; 
  }
}

template<typename T, typename Tg>
void af(const T& x, T& ax, const T& h, T& ah, const Tg& g, T& ag, T& y, T& ay, bool adjoint=false) {
  if (!adjoint) 
    y=x+h+g; 
  else {
    ax+=ay; ah+=ay; ag+=ay; ay=0;
  }
}

template<typename T>
void aF(const T& x, T& ax, T& y, T& ay) {
  T h_, ah_=0; ah(x,ax,h_,ah_);
  T g_, ag_=0; ag(x,ax,h_,ah_,g_,ag_);
  af(x,ax,h_,ah_,g_,ag_,y,ay);
  af(x,ax,h_,ah_,g_,ag_,y,ay,true);
  ag(x,ax,h_,ah_,g_,ag_,true);
  ah(x,ax,h_,ah_,true);
}

int main() {
  double x=1.5, ax=0, y, ay=1;
  aF(x,ax,y,ay);
  cout << y << ' ' << ax << endl;
  return 0;
}

